"""Shared decorators and exceptions used in virtual tiles and devices."""

import struct
import binascii
import inspect
from iotile.core.exceptions import IOTileException


class RPCNotFoundError(IOTileException):
    """Exception thrown when an RPC is not found."""
    pass


class RPCInvalidArgumentsError(IOTileException):
    """Exception thrown when an RPC with a fixed parameter format has invalid arguments."""
    pass


class RPCInvalidReturnValueError(IOTileException):
    """Exception thrown when the return value of an RPC does not conform to its known format."""
    pass


class RPCInvalidIDError(IOTileException):
    """Exception thrown when an RPC is created with an invalid RPC id."""
    pass


class TileNotFoundError(IOTileException):
    """Exception thrown when an RPC is sent to a tile that does not exist."""
    pass

class RPCErrorCode(IOTileException):
    """Exception thrown from an RPC implementation to set the status code."""

    def __init__(self, status_code):
        super(RPCErrorCode, self).__init__("RPC returned application defined status code %d" % status_code, code=status_code)


def _create_argcode(code, arg_bytes):
    if not code.endswith('V'):
        return "<" + code

    code = code[:-1]
    fixed_size = struct.calcsize("<" + code)
    var_size = len(arg_bytes) - fixed_size

    if var_size < 0:
        raise RPCInvalidArgumentsError("Argument was too small for variable size argument value", arg_format=code,
                                       minimum_size=fixed_size, actual_size=len(arg_bytes),
                                       payload=binascii.hexlify(arg_bytes))

    return "<" + code + "%ds" % var_size


def _create_respcode(code, resp):
    if not code.endswith('V'):
        return "<" + code

    code = code[:-1]

    final_length = len(resp[-1])
    fixed_size = struct.calcsize("<" + code)

    if fixed_size + final_length > 20:
        raise RPCInvalidReturnValueError("Variable length return value is too large for rpc response payload (20 bytes)",
                                         fixed_code=code, fixed_length=fixed_size, variable_length=final_length,
                                         variable_payload=binascii.hexlify(resp[-1]))
    return "<" + code + "%ds" % final_length


def pack_rpc_payload(arg_format, args):
    """Pack an RPC payload according to arg_format.

    Args:
        arg_format (str): a struct format code (without the <) for the
            parameter format for this RPC.  This format code may include the final
            character V, which means that it expects a variable length bytearray.
        args (list): A list of arguments to pack according to arg_format.

    Returns:
        bytes: The packed argument buffer.
    """

    code = _create_respcode(arg_format, args)
    return struct.pack(code, *args)


def unpack_rpc_payload(resp_format, payload):
    """Unpack an RPC payload according to resp_format.

    Args:
        resp_format (str): a struct format code (without the <) for the
            parameter format for this RPC.  This format code may include the final
            character V, which means that it expects a variable length bytearray.
        payload (bytes): The binary payload that should be unpacked.

    Returns:
        list: A list of the unpacked payload items.
    """

    code = _create_argcode(resp_format, payload)
    return struct.unpack(code, payload)


def rpc(address, rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID and address.

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        address (int): The address of the mock tile this RPC is for
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC.  This format code may include the final
            character V, which means that it expects a variable length bytearray.
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC. This format code may include the final
            character V, which means that it expects a variable length bytearray.
    """

    if rpc_id < 0 or rpc_id > 0xFFFF:
        raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

    def _rpc_wrapper(func):
        def _rpc_executor(self, payload):
            try:
                args = unpack_rpc_payload(arg_format, payload)
            except struct.error as exc:
                raise RPCInvalidArgumentsError(str(exc), arg_format=arg_format, payload=binascii.hexlify(payload))

            resp = func(self, *args)

            if resp is None:
                resp = []

            if resp_format is not None:
                try:
                    return pack_rpc_payload(resp_format, resp)
                except struct.error as exc:
                    raise RPCInvalidReturnValueError(str(exc), resp_format=resp_format, resp=repr(resp))

            return resp

        _rpc_executor.rpc_id = rpc_id
        _rpc_executor.rpc_addr = address
        _rpc_executor.is_rpc = True
        return _rpc_executor

    return _rpc_wrapper


def tile_rpc(rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID on a tile.

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC
    """

    return rpc(None, rpc_id, arg_format, resp_format)


class RPCDispatcher(object):
    """A simple dispatcher that can store and call RPCs."""

    def __init__(self, *args, **kwargs):
        super(RPCDispatcher, self).__init__(*args, **kwargs)
        self._rpcs = {}

        # Add any RPCs defined using decorators on this class
        for _name, value in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(value, 'is_rpc'):
                self.add_rpc(value.rpc_id, value)

    def add_rpc(self, rpc_id, func):
        """Add an RPC.

        Args:
            rpc_id (int): The ID of the RPC
            func (callable): The RPC implementation.
                The signature of callable should be callable(args) taking
                a bytes object with the argument and returning a bytes object
                with the response.
        """

        self._rpcs[rpc_id] = func

    def has_rpc(self, rpc_id):
        """Check if an RPC is defined.

        Args:
            rpc_id (int): The RPC to check

        Returns:
            bool: Whether it exists
        """

        return rpc_id in self._rpcs

    def call_rpc(self, rpc_id, payload=bytes()):
        """Call an RPC by its ID.

        Args:
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes

        Returns:
            str: The response payload from the RPC
        """
        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if rpc_id not in self._rpcs:
            raise RPCNotFoundError("rpc_id: {}".format(rpc_id))

        return self._rpcs[rpc_id](payload)



"""A sparse memory map for debugging purposes
"""

from __future__ import unicode_literals
from builtins import range
from collections import namedtuple
import binascii
import string
from iotile.core.exceptions import ArgumentError


MemorySegment = namedtuple('MemorySegment', ['start_address', 'end_address', 'length', 'data'])
DisjointSegment = namedtuple('DisjointSegment', [])
OverlappingSegment = namedtuple('OverlappingSegment', [])

class SparseMemory(object):
    """A sparse memory map for debugging purposes

    You can add memory segments into the memory map
    a little bit at a time and decode any section into
    a python object using registered decoder functions.
    """
    def __init__(self):
        self._segments = []

    def add_segment(self, address, data, overwrite=False):
        """Add a contiguous segment of data to this memory map

        If the segment overlaps with a segment already added , an
        ArgumentError is raised unless the overwrite flag is True.

        Params:
            address (int): The starting address for this segment
            data (bytearray): The data to add
            overwrite (bool): Overwrite data if this segment overlaps
                with one previously added.
        """

        seg_type = self._classify_segment(address, len(data))
        if not isinstance(seg_type, DisjointSegment):
            raise ArgumentError("Unsupported segment type")

        segment = MemorySegment(address, address+len(data)-1, len(data), bytearray(data))
        self._segments.append(segment)

    def _create_slice(self, key):
        """Create a slice in a memory segment corresponding to a key."""

        if isinstance(key, slice):
            step = key.step
            if step is None:
                step = 1

            if step != 1:
                raise ArgumentError("You cannot slice with a step that is not equal to 1", step=key.step)

            start_address = key.start
            end_address = key.stop - 1

            start_i, start_seg = self._find_address(start_address)
            end_i, _end_seg = self._find_address(end_address)

            if start_seg is None or start_i != end_i:
                raise ArgumentError("Slice would span invalid data in memory", start_address=start_address, end_address=end_address)

            block_offset = start_address - start_seg.start_address
            block_length = end_address - start_address + 1

            return start_seg, block_offset, block_offset + block_length
        elif isinstance(key, int):
            start_i, start_seg = self._find_address(key)
            if start_seg is None:
                raise ArgumentError("Requested invalid address", address=key)

            return start_seg, key - start_seg.start_address, None
        else:
            raise ArgumentError("Unknown type of address key", address=key)

    def __getitem__(self, key):
        seg, start, end = self._create_slice(key)

        if end is None:
            return seg.data[start]

        return seg.data[start:end]

    def __setitem__(self, key, item):
        seg, start, end = self._create_slice(key)

        if end is None:
            seg.data[start] = item
        else:
            seg.data[start:end] = item

    @classmethod
    def _in_segment(cls, address, segment):
        return address >= segment.start_address and address <= segment.end_address

    def _find_address(self, address):
        for i, segment in enumerate(self._segments):
            if self._in_segment(address, segment):
                return i, segment

        return -1, None

    def _classify_segment(self, address, length):
        """Determine how a new data segment fits into our existing world

        Params:
            address (int): The address we wish to classify
            length (int): The length of the segment

        Returns:
            int: One of SparseMemoryMap.prepended
        """

        end_address = address + length - 1

        _, start_seg = self._find_address(address)
        _, end_seg = self._find_address(end_address)

        if start_seg is not None or end_seg is not None:
            raise ArgumentError("Overlapping segments are not yet supported", address=address, length=length)

        return DisjointSegment()

    @classmethod
    def _iter_groups(cls, data, chunk_length):
        for i in range(0, len(data), chunk_length):
            yield data[i:i+chunk_length]

    @classmethod
    def _format_line(cls, start_address, data):
        hexdata = binascii.hexlify(data).decode('utf-8')
        spaced_hex = ' '.join([hexdata[i:i+2] for i in range(0, len(hexdata), 2)])
        separated_hex = spaced_hex[0:len(spaced_hex)//2] + ' ' + spaced_hex[len(spaced_hex)//2:]

        asciidata = ''

        for x in data:
            if chr(x) in string.printable and chr(x) != '\n' and chr(x) != '\r' and chr(x) != '\t':
                asciidata += chr(x)
            else:
                asciidata += '.'

        return "0x{:08x}: {}  {}".format(start_address, separated_hex, asciidata)

    def __str__(self):
        """Convert to string as a 64 byte wide hex dump
        """

        stringdata = ""

        for segment in self._segments:
            # Insert a ... between every distinct segment
            if len(stringdata) > 0:
                stringdata += ("----- "*13) + "\n"

            current_addr = segment.start_address
            for linedata in self._iter_groups(segment.data, 16):
                line = self._format_line(current_addr, linedata)
                stringdata += line + '\n'
                current_addr += 16

        return stringdata



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

# IOTile Exceptions

from typedargs.exceptions import KeyValueException as IOTileException
from typedargs.exceptions import ArgumentError, ValidationError, ConversionError, NotFoundError, TypeSystemError, InternalError


class TimeoutExpiredError(IOTileException):
    """
    The method timed out, usually indicating that a communication failure
    occurred, either with another process or with a hardware module.
    """

    pass


class DataError(IOTileException):
    """
    The method relied on data pass in by the user and the data was invalid.

    This could be because a file was the wrong type or because a data provider
    returned an unexpected result.  The parameters passed with this exception
    provide more detail on what occurred and where.
    """

    pass

class InternalError(IOTileException):
    """
    The method could not be completed with the user input passed for
    an unexpected reason.  This does not signify a bug in the API
    method code.  More details should be passed in the arguments.
    """

    pass

class APIError(IOTileException):
    """
    An internal API error occured during the execution of the method.
    This should only be returned if the error was unforeseen and not
    caused in any way by user input.  If the problem is that a user
    input is invalid for the API call, ValidationError should be
    thrown instead.

    All instances of APIError being thrown are bugs that should be
    reported and fixed.
    """

    pass

class BuildError(IOTileException):
    """
    There is an error with some part of the build system.  This does not
    mean that there is a compilation error but rather that a required part
    of the build process did not complete successfully.  This exception means
    that something is misconfigured.
    """

    pass


class ExternalError(IOTileException):
    """
    The external environment is not properly configured for the IOTile API command that was called.
    This can be because a required program was not installed or accessible or because
    a required environment variable was not defined.
    """

    pass

class HardwareError(IOTileException):
    """
    There was an issue communicating with or controlling an IOTile hardware module.  This
    exception anchors a range of exceptions that refer to specific kinds of hardware issues.

    By catching this exception, you will catch any sort of hardware failure.  If you are
    interested in specific kinds of hardware errors, you can look for or catch subclasses
    of this exception.
    """

    pass
